package cerebral.layout;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

import javax.swing.ProgressMonitor;
import javax.swing.SwingUtilities;

import prefuse.action.Action;
import prefuse.data.Edge;
import prefuse.data.Node;
import prefuse.data.tuple.TupleSet;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualGraph;
import prefuse.visual.VisualItem;
import cerebral.Cerebral;
import cerebral.CerebralLayout;
import cerebral.EdgeBundler;

public class SearchGridLayout extends Action implements CerebralLayout {

	private static final int COOLING_PHASES = 30;

	private static final int NODE_EDGE_PENALTY = 9;

	private static final int EDGE_PENALTY = 3;

	private int gridWidth;

	private int gridHeight;

	int iteration;

	private List[] candidatePoints;

	private Node[][] nodeAtPosition;

	private int[] nodeCounts;

	private int[] layerBounds;

	private SpaceGrid spaceGrid;

	public static int SCALE = 60;

	private String cancelAction;

	private Point neighbPoint = new Point();

	private Random generator = new Random();

	private Point[] nodepos;

	private Point nodeOrigSave = new Point();

	private boolean firstRun;

	private VisualGraph graph;

	private int numLayers;

	private double temperature;

	private boolean quitEarly;

	private ProgressMonitor progressMonitor;

	private long startTime;

	private List moveableNodes = new ArrayList();

	private int energy;

	// For grid scoring evaluation
	public SearchGridLayout(VisualGraph graph, String cancelAction,
			int numLayers) {
		this.cancelAction = cancelAction;
		this.graph = graph;
		this.numLayers = numLayers;
		firstRun = true;

	}

	public boolean pointOccupied(int x, int y) {
		return nodeAtPosition[x / SCALE][y / SCALE] != null;
	}

	public NodeItem getNodeAtPosition(int x, int y) {
		return (NodeItem) nodeAtPosition[x / SCALE][y / SCALE];

	}

	private void checkNodeAtPos() {
		Set nodesSeen = new HashSet();
		for (int x = 0; x < nodeAtPosition.length; x++) {
			for (int y = 0; y < nodeAtPosition[0].length; y++) {
				Node n = nodeAtPosition[x][y];
				if (n != null) {
					if (nodesSeen.contains(n)) {
						System.out.println("NODESSEEN inconsisten");
					} else {
						nodesSeen.add(n);
					}
				}
			}
		}

		Set uniqueNodeLoc = new HashSet();

		for (int i = 0; i < nodepos.length; i++) {
			Point p = nodepos[i];
			if (p != null) {
				Integer hash = new Integer(p.x * 10000 + p.y);
				if (hash.intValue() != 0) {
					if (uniqueNodeLoc.contains(hash)) {
						System.out
								.println("nodepos contains two different points that map to the same location");

					}
					uniqueNodeLoc.add(hash);
				}
			}
		}
	}

	public void initialize() {

		startTime = System.currentTimeMillis();
		System.out.println("Starting layout.");
		nodepos = new Point[graph.getNodeCount() + 1];
		int rootV = (int) Math.sqrt(graph.getNodeCount());
		gridWidth = (int) (2.5 * rootV);

		for (int i = 0; i < nodepos.length; i++) {
			nodepos[i] = new Point();
		}
		iteration = 1;
		// add 1 for the downstream layer
		nodeCounts = new int[numLayers + 1];
		Arrays.fill(nodeCounts, 0);

		TupleSet nodes = graph.getNodes();

		int anyLayerCount = 0;
		Iterator nodesIt = nodes.tuples();
		while (nodesIt.hasNext()) {
			NodeItem node = (NodeItem) nodesIt.next();
			node.setStartX(node.getX());
			node.setStartY(node.getY());

			if (node.getBoolean(Cerebral.ANY_LAYER)
					&& !node.getBoolean(Cerebral.DOWNSTREAM_GENE)) {
				anyLayerCount++;
			} else {
				if (node.getBoolean(Cerebral.DOWNSTREAM_GENE)) {
					nodeCounts[numLayers]++;
				} else {
					int bioCategory = node.getInt(Cerebral.BIO_CATEGORY);
					nodeCounts[bioCategory]++;
				}
			}
		}

		layerBounds = new int[numLayers + 1];
		int bottom = 0;
		int top = 0;

		float avgAnyLayersPerLayer = anyLayerCount / (numLayers + 1);
		for (int i = 0; i < numLayers + 1; i++) {
			bottom = top;

			top += (int) (((nodeCounts[i] + avgAnyLayersPerLayer) / (float) nodes
					.getTupleCount())
					* (3.0 / 4.0) * gridWidth);
			if (top - bottom == 0) {
				top++;
			}
			if (top - bottom == 1) {
				top++;
			}
			if (top - bottom == 2) {
				top++;
			}
			while ((top - bottom) * gridWidth < 2 * (nodeCounts[i] + avgAnyLayersPerLayer)) {
				top++;
			}
			layerBounds[i] = top;
		}
		gridHeight = top;
		System.out.println("Using a grid size of [" + gridWidth + ":"
				+ gridHeight + "]");

		candidatePoints = new ArrayList[numLayers + 1];
		for (int i = 0; i < numLayers + 1; i++) {
			candidatePoints[i] = new ArrayList();
		}

		bottom = 0;
		top = 0;
		for (int i = 0; i < nodeCounts.length; i++) {
			bottom = top;
			top = layerBounds[i];
			for (int j = 0; j < gridWidth; j++) {
				for (int k = bottom; k < top; k++) {
					Point p = new Point(j, k);
					candidatePoints[i].add((p));
				}
			}
		}

		nodesIt = nodes.tuples();
		int maxGridPoint = Integer.MIN_VALUE;
		while (nodesIt.hasNext()) {
			NodeItem node = (NodeItem) nodesIt.next();
			int x = (int) node.getEndX();
			if (x / SCALE > maxGridPoint) {
				maxGridPoint = x / SCALE;
			}
			int nx = nodepos[node.getRow()].x;
			if (nx > maxGridPoint) {
				maxGridPoint = nx;
			}
		}
		maxGridPoint = (maxGridPoint > gridWidth) ? maxGridPoint : gridWidth;

		initializeNodeAtPosition(maxGridPoint,
				layerBounds[layerBounds.length - 1]);

		assignMoveableNodes();
		assignRandomValidPositions();

		initializeSpaceGrid();

		// We want the initial temperature to be the temperature of our current
		// layout
		energy = 0;
		nodesIt = nodes.tuples();
		while (nodesIt.hasNext()) {
			NodeItem node = (NodeItem) nodesIt.next();
			energy += scoreNode(node);
		}
		if (nodes.getTupleCount() > 0) {
			temperature = energy / nodes.getTupleCount();
			quitEarly = false;
		} else {
			quitEarly = true;
		}
	}

	public void run(double frac) {

		if (firstRun) {
			initialize();
			firstRun = false;
		}

		Cerebral cytoPrefuse = (Cerebral) m_vis.getDisplay(0);

		improveLayout();

		if (iteration == COOLING_PHASES || quitEarly) {
			System.out.println("Layout time. "
					+ (System.currentTimeMillis() - startTime));
			System.out.println("Layout finished.  Total iterations: "
					+ iteration);
			SwingUtilities.invokeLater(new Runnable() {

				public void run() {
					getProgressMonitor().close();
				}

			});

			// We're done.
			m_vis.cancel(cancelAction);

			EdgeBundler edgeBundler = cytoPrefuse.edgeBundler;
			if (edgeBundler != null) {
				edgeBundler.recomputeLocations();
			}

			cytoPrefuse.viewCoordinator.positionDividers();
			cytoPrefuse.setCytoscapePostionsFromLayout(graph.getNodes());
			cytoPrefuse.viewCoordinator.run("animateNodes");
			cytoPrefuse.viewCoordinator.fitViewToBounds(1000);

			m_vis.invalidateAll();
			m_vis.run("repaint");
			firstRun = true;

		} else {
			temperature = temperature * 0.60;
		}
	}

	private void assignMoveableNodes() {
		moveableNodes.clear();
		TupleSet nodes = graph.getNodes();
		Iterator nodesIt = nodes.tuples();
		while (nodesIt.hasNext()) {
			NodeItem node = (NodeItem) nodesIt.next();
			if (!node.getBoolean(Cerebral.PINNED)) {
				moveableNodes.add(node);
			} else {

				int x = (int) node.getEndX();
				int y = (int) node.getEndY();

				Point p = nodepos[node.getRow()];
				p.x = x / SCALE;
				p.y = y / SCALE;

				if (p.x < 0) {
					p.x = 0;
				}
				if (p.x > nodeAtPosition.length - 1) {
					p.x = nodeAtPosition.length - 1;
				}
				if (p.y < 0) {
					p.y = 0;
				}
				if (p.y > nodeAtPosition[0].length - 1) {
					p.y = nodeAtPosition[0].length - 1;
				}
				nodepos[node.getRow()] = p;
				nodeAtPosition[p.x][p.y] = node;
			}
		}
	}

	private void initializeSpaceGrid() {

		// the spacegrid needs to be big enough for any node to be tracked,
		// not just the nodes we're laying out (for the nodes we are laying
		// out could join to nodes outside the region.)
		int maxX = Integer.MIN_VALUE;
		int maxY = Integer.MIN_VALUE;
		TupleSet nodes = graph.getNodes();
		Iterator nodesIt = nodes.tuples();
		while (nodesIt.hasNext()) {
			NodeItem node = (NodeItem) nodesIt.next();
			Point p = nodepos[node.getRow()];
			if (p.x > maxX) {
				maxX = p.x;
			}
			if (p.y > maxY) {
				maxY = p.y;
			}
		}
		int width = maxX > gridWidth ? maxX : gridWidth;
		int height = maxY > gridHeight ? maxY : gridHeight;
		spaceGrid = new SpaceGrid(width + 1, height + 1, graph.getNodeCount(),
				this);

		Iterator iter = graph.getNodes().tuples();
		Set edgesAdded = new HashSet();
		while (iter.hasNext()) {
			NodeItem node = (NodeItem) iter.next();
			spaceGrid.addNode(node);
			Iterator edgesIt = node.edges();
			while (edgesIt.hasNext()) {
				Edge edge = (Edge) edgesIt.next();
				if (!edgesAdded.contains(edge)) {
					spaceGrid.addEdge(edge);
					edgesAdded.add(edge);
				}
			}
		}
	}

	private void initializeNodeAtPosition(int x, int y) {
		nodeAtPosition = new Node[x + 1][y + 1];
		for (int xi = 0; xi <= x; xi++) {
			Arrays.fill(nodeAtPosition[xi], null);
		}
	}

	public final Point getPositionOfNode(Node n) {
		return nodepos[n.getRow()];
	}

	private void improveLayout() {

		final long startTime = System.currentTimeMillis();
		iteration++;
		int escapeMinima = 0;

		for (int i = 0; i < 50 * moveableNodes.size(); i++) {

			// pick a node at random
			NodeItem node = (NodeItem) moveableNodes.get(generator
					.nextInt(moveableNodes.size()));

			Point trialPoint;
			do {
				// and a random spot in the grid
				int layer = getLayer(generator, node);

				int sample = generator.nextInt(candidatePoints[layer].size());
				trialPoint = (Point) candidatePoints[layer].get(sample);
			} while (!pointIsEmpty(trialPoint));

			int improvement = 0;
			removeNodeFromSpaceGrid(node);
			int oldScore = scoreNode(node);
			nodeOrigSave.x = nodepos[node.getRow()].x;
			nodeOrigSave.y = nodepos[node.getRow()].y;

			moveNodeToEmptyPoint(node, trialPoint, false);
			improvement = oldScore - scoreNode(node);
			boolean acceptChange = false;
			if (improvement > 0) {
				acceptChange = true;
			} else {
				// SImulated annealing test
				if (Math.exp(improvement / temperature) > Math.random()) {
					// System.out.println ("Ratio " + (improvement/temperature)
					// + " :exp " +Math.exp (improvement/temperature));
					acceptChange = true;
					escapeMinima++;
				}
			}
			if (acceptChange) {
				energy -= improvement;
				setDisplayFromNodepos(node);
			} else {
				moveNodeToEmptyPoint(node, nodeOrigSave, false);
			}
			addNodeToSpaceGrid(node);

			if ((i % 100) == 0) {
				if (getProgressMonitor().isCanceled()) {
					quitEarly = true;
					break;
				}
			}
		}

		// System.out.print(energy + ", ");
		SwingUtilities.invokeLater(new Runnable() {

			public void run() {
				getProgressMonitor().setProgress(
						(int) (iteration / ((float) COOLING_PHASES) * 1000));
				String message = "Estimated time left ";
				long estimatedTimeLeft = (COOLING_PHASES - iteration)
						* (System.currentTimeMillis() - startTime) / 1000;
				if (estimatedTimeLeft > 120) {
					message += estimatedTimeLeft / 60 + " minutes.";
				} else {
					message += ((estimatedTimeLeft / 5) + 1) * 5 + " seconds.";
				}
				getProgressMonitor().setNote(message);
			}

		});

	}

	private void addNodeToSpaceGrid(Node node) {
		spaceGrid.addNode((NodeItem) node);
		Iterator edges = node.edges();
		while (edges.hasNext()) {
			Edge e = (Edge) edges.next();
			spaceGrid.addEdge(e);
		}
	}

	private void removeNodeFromSpaceGrid(Node node) {
		Iterator edges = node.edges();
		while (edges.hasNext()) {
			Edge e = (Edge) edges.next();
			spaceGrid.deleteEdge(e);
		}
		spaceGrid.deleteNode((NodeItem) node);
	}

	private int scoreNode(NodeItem node) {
		int score = 0;

		if (node.getBoolean(Cerebral.DOWNSTREAM_GENE)) {
			score = countCrossings(node);
			score += countNeighbourNodeDistance(node);
			score += 10 * NODE_EDGE_PENALTY
					* countFunctionNeighbourNodeDistance(node);
			score += 10 * NODE_EDGE_PENALTY
					* countOtherFunctionInNeighbourhood(node);
		} else {
			score = countCrossings(node);
			score += countNeighbourNodeDistance(node);
			// score += countFunctionNeighbourNodeDistance(node);
		}

		return score;
	}

	private int countNeighbourNodeDistance(NodeItem node, Iterator neighbs) {
		int score = 0;
		Point nodePos = nodepos[node.getRow()];
		while (neighbs.hasNext()) {
			Node neighbour = (Node) neighbs.next();
			Point neighbPos = nodepos[neighbour.getRow()];
			int increment = Math.abs(neighbPos.x - nodePos.x)
					+ Math.abs(neighbPos.y - nodePos.y);
			score += increment;
		}
		return score;
	}

	private int countFunctionNeighbourNodeDistance(NodeItem node) {

		Vector functionNeighbourList = (Vector) node
				.get(Cerebral.FUNCTION_NEIGHBOURS);
		int score = 0;
		if (functionNeighbourList != null) {
			score = countNeighbourNodeDistance(node, functionNeighbourList
					.iterator());
		}
		return score;
	}

	private int countOtherFunctionInNeighbourhood(NodeItem node) {
		int score = 0;
		boolean functionUnknown = node
				.getBoolean(Cerebral.BIO_FUNCTION_UNKNOWN);
		int bioFunction = node.getInt(Cerebral.BIO_FUNCTION);
		Point pos = nodepos[node.getRow()];
		for (int x = -1; x <= 1; x++) {
			for (int y = -1; y <= 1; y++) {
				neighbPoint.x = pos.x + x;
				neighbPoint.y = pos.y + y;
				if (neighbPoint.x >= 0 && neighbPoint.x < nodeAtPosition.length
						&& neighbPoint.y >= 0
						&& neighbPoint.y < nodeAtPosition[0].length
						&& (x != 0 || y != 0)) {
					Node neighbour = nodeAtPosition[neighbPoint.x][neighbPoint.y];
					if (neighbour != null) {
						int neighbBioFunction = neighbour
								.getInt(Cerebral.BIO_FUNCTION);
						boolean neighbFunctionUnknown = neighbour
								.getBoolean(Cerebral.BIO_FUNCTION_UNKNOWN);
						if (!functionUnknown && !neighbFunctionUnknown) {
							if (neighbBioFunction != bioFunction) {
								score++;
							}
						} else if (functionUnknown ^ neighbFunctionUnknown) {
							score++;
						}
					}
				}
			}
		}
		return score;

	}

	private int countNeighbourNodeDistance(NodeItem node) {
		return countNeighbourNodeDistance(node, node.neighbors());
	}

	private int countCrossings(NodeItem node) {
		int edgeCrossings = 0;
		Iterator edgesIt = node.edges();
		while (edgesIt.hasNext()) {
			Edge e = (Edge) edgesIt.next();
			edgeCrossings += spaceGrid.countCrossings(e, EDGE_PENALTY,
					NODE_EDGE_PENALTY);
		}

		int thisNodeEdgeCrossings = spaceGrid.countNodeEdgeCrossingsAtCellP(nodepos[node
				.getRow()], NODE_EDGE_PENALTY);
		return edgeCrossings + thisNodeEdgeCrossings;

	}

	private void assignRandomValidPositions() {
		Iterator nodesIt = moveableNodes.iterator();
		while (nodesIt.hasNext()) {
			NodeItem node = (NodeItem) nodesIt.next();
			int layer = getLayer(generator, node);
			int randPos;
			Point pCandidate;
			do {
				randPos = generator.nextInt(candidatePoints[layer].size());
				pCandidate = (Point) candidatePoints[layer].get(randPos);
			} while (!pointIsEmpty(pCandidate));

			Point p = new Point(pCandidate.x, pCandidate.y);
			nodepos[node.getRow()] = p;
			nodeAtPosition[p.x][p.y] = node;
			setDisplayFromNodepos(node);
		}
	}

	private void setDisplayFromNodepos(NodeItem node) {
		Point p = nodepos[node.getRow()];
		// node.setX(p.x*SCALE);
		// node.setY(p.y *SCALE);
		node.setEndX(p.x * SCALE);
		node.setEndY(p.y * SCALE);
	}

	private void moveNodeToEmptyPoint(Node node, Point p, boolean updateGrid) {
		if (updateGrid)
			removeNodeFromSpaceGrid(node);

		Point oldPos = nodepos[node.getRow()];
		nodeAtPosition[oldPos.x][oldPos.y] = null;

		oldPos.x = p.x;
		oldPos.y = p.y;

		nodeAtPosition[p.x][p.y] = node;
		if (updateGrid)
			addNodeToSpaceGrid(node);
	}

	private boolean pointIsEmpty(Point p) {
		if (nodeAtPosition[p.x][p.y] != null)
			return false;
		else {
			// make sure it's not a layer boundary
			for (int i = 0; i < layerBounds.length; i++) {
				if (p.y == layerBounds[i])
					return false;
			}
			return true;
		}
	}

	private int getLayer(Random generator, Node node) {

		int layer = 0;
		if (node.getBoolean(Cerebral.DOWNSTREAM_GENE)) {
			layer = numLayers;
		} else {
			layer = node.getInt(Cerebral.BIO_CATEGORY);
		}
		boolean anyLayer = node.getBoolean(Cerebral.ANY_LAYER)
				&& !node.getBoolean(Cerebral.DOWNSTREAM_GENE);
		if (anyLayer) {
			layer = generator.nextInt(numLayers + 1);
		}
		return layer;
	}

	public Point pointLocation(VisualItem item) {
		return nodepos[item.getRow()];
	}

	public void setNodeAtLocation(VisualItem item, int unscaledx, int unscaledy) {
		Point scaledPoint = new Point(unscaledx / SCALE, unscaledy / SCALE);
		moveNodeToEmptyPoint((Node) item, scaledPoint, false);
	}

	public void setProgressMonitor(ProgressMonitor progressMonitor) {
		this.progressMonitor = progressMonitor;
	}

	public ProgressMonitor getProgressMonitor() {
		return progressMonitor;
	}

}