package cerebral.layout;

import java.awt.Point;
import java.util.Arrays;

import cerebral.CerebralLayout;

import prefuse.data.Edge;
import prefuse.data.Node;
import prefuse.visual.NodeItem;

public class SpaceGrid {
    private interface GridVisitor {
        void visitGridPoint(int x, int y, int edgePenalty, int nodeEdgePenalty,
                boolean nearCenter, Node edgeSource, Node edgeDest);
    }

    private class OccupyGrid implements GridVisitor {

        public void visitGridPoint(int x, int y, int edgePenalty,
                int nodeEdgePenalty, boolean nearCenter, Node edgeSource,
                Node edgeDest) {
            if (nearCenter) {
                edgesThroughCenterCount[x][y]++;
            }
            edgesInCellCount[x][y]++;
        }
    }

    private class VacateGrid implements GridVisitor {

        public void visitGridPoint(int x, int y, int edgePenalty,
                int nodeEdgePenalty, boolean nearCenter, Node edgeSource,
                Node edgeDest) {
            if (nearCenter) {
                edgesThroughCenterCount[x][y]--;
            }
            edgesInCellCount[x][y]--;
        }
    }

    private class CountIntersections implements GridVisitor {
        public void visitGridPoint(int x, int y, int edgePenalty,
                int nodeEdgePenalty, boolean nearCenter, Node edgeSource,
                Node edgeDest) {

            crossingScore += edgesInCellCount[x][y] * edgePenalty;
            if (nearCenter) {
                if (occupiedNodes[x][y]) {
                    crossingScore += nodeEdgePenalty;
                }
            }

        }

    }

    private boolean[][] occupiedNodes;

    private int[][] edgesThroughCenterCount;

    private int[][] edgesInCellCount;

    private OccupyGrid occupyGrid = new OccupyGrid();

    private VacateGrid vacateGrid = new VacateGrid();

    private CountIntersections countIntersections = new CountIntersections();

    private int crossingScore;

    private CerebralLayout layout;

    SpaceGrid(int x, int y, int maxNodes, CerebralLayout inCerebralLayout) {
        occupiedNodes = new boolean[x][y];
        edgesThroughCenterCount = new int[x][y];
        edgesInCellCount = new int[x][y];

        for (int xi = 0; xi < x; xi++) {
            Arrays.fill(edgesInCellCount[xi], 0);
            Arrays.fill(edgesThroughCenterCount[xi], 0);
            Arrays.fill(occupiedNodes[xi], false);
        }
        

        layout = inCerebralLayout;
    }

    public void checkHealth() {
        // For debugging
        for (int x = 0; x < edgesInCellCount.length; x++) {
            for (int y = 0; y < edgesInCellCount[x].length; y++) {
                if (edgesInCellCount[x][y] < 0) {
                    System.out.println("[" + x + ":" + y + "] has score "
                            + edgesInCellCount[x][y]);
                }
            }
        }
    }

    public void addEdge(Edge e) {
        visitGridsCrossedByEdge(e, occupyGrid, 0, 0);
    }

    public void deleteEdge(Edge e) {
        visitGridsCrossedByEdge(e, vacateGrid, 0, 0);

    }

    private void visitGridsCrossedByEdge(Edge e, GridVisitor gv,
            int edgePenalty, int nodeEdgePenalty) {
        Point p0 = layout.getPositionOfNode(e.getSourceNode());
        Point p1 = layout.getPositionOfNode(e.getTargetNode());

        Node edgeSource = e.getSourceNode();
        Node edgeDest = e.getTargetNode();
        if (edgeSource.getRow() != edgeDest.getRow()) {

            int x0 = p0.x;
            int x1 = p1.x;
            int y0 = p0.y;
            int y1 = p1.y;

            boolean steep = Math.abs(y1 - y0) > Math.abs(x1 - x0);
            if (steep) {
                int temp = x0;
                x0 = y0;
                y0 = temp;

                temp = x1;
                x1 = y1;
                y1 = temp;

            }
            if (x0 > x1) {

                int temp = x0;
                x0 = x1;
                x1 = temp;

                temp = y0;
                y0 = y1;
                y1 = temp;

            }
            int deltax = x1 - x0;
            int deltay = Math.abs(y1 - y0);
            double error = 0;
            int y = y0;
            int ystep = (y0 < y1) ? 1 : -1;
            double deltaErr = (deltay) / (double) deltax;

            for (int x = x0; x <= x1; x++) {
                if (steep)
                    gv.visitGridPoint(y, x, edgePenalty, nodeEdgePenalty, Math
                            .abs(error) < 0.3, edgeSource, edgeDest);
                else
                    gv.visitGridPoint(x, y, edgePenalty, nodeEdgePenalty, Math
                            .abs(error) < 0.3, edgeSource, edgeDest);

                error += deltaErr;
                if (error > 0.5) {
                    // Here's the modified version of Bresenham's. We also visit
                    // the node we
                    // passed through as we change the y co-ordinate.
                    if (x < x1) {
                        if (error > 0.9999 && error < 1.0001) {

                            // we're exiting at an exact corner.
                            y = y + ystep;
                        } else if (error > 1.0) {

                            // we're over, check the new y immediately
                            y = y + ystep;
                            if (steep)
                                gv.visitGridPoint(y, x, edgePenalty,
                                        nodeEdgePenalty, false, edgeSource,
                                        edgeDest);
                            else
                                gv.visitGridPoint(x, y, edgePenalty,
                                        nodeEdgePenalty, false, edgeSource,
                                        edgeDest);
                        } else {
                            // we're under, check one more x
                            if (steep)
                                gv.visitGridPoint(y, x + 1, edgePenalty,
                                        nodeEdgePenalty, false, edgeSource,
                                        edgeDest);
                            else
                                gv.visitGridPoint(x + 1, y, edgePenalty,
                                        nodeEdgePenalty, false, edgeSource,
                                        edgeDest);
                            y = y + ystep;
                        }
                        error -= 1.0;
                    }
                }
            }
        }
    }

    public void addNode(NodeItem n) {
        Point nPos = layout.getPositionOfNode(n);
        occupiedNodes[nPos.x][nPos.y] = true;

    }

    public void deleteNode(NodeItem n) {
        Point nPos = layout.getPositionOfNode(n);
        occupiedNodes[nPos.x][nPos.y] = false;
    }

    public int countCrossings(Edge e, int edgePenalty, int nodeEdgePenalty) {
        crossingScore = 0;
        visitGridsCrossedByEdge(e, countIntersections, edgePenalty,
                nodeEdgePenalty);
        return crossingScore;

    }

    public int countNodeEdgeCrossingsAtCellP(Point p, int nodeEdgePenalty) {
        int x = p.x;
        int y = p.y;

        int crossings = 0;
        crossings += edgesThroughCenterCount[x][y] * nodeEdgePenalty;
        return crossings;
    }
}
